/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="ExtTrigConditionDataCV"> ********
Author: Rahkola, R
$Id: ExtTrigConditionData.c,v 1.4 2002/04/30 17:25:40 rrahkola Exp $
********* </lalVerbatim> ********/

/* Include input/output functions if we're in the debug stage */
#ifdef DEBUG_EXTTRIG
#include <stdio.h>
#include <lal/PrintFTSeries.h>
#endif // DEBUG_EXTTRIG

/* standard C functions */
#include <math.h>
#include <string.h>
#include <mpi.h>

/* standard LAL functions*/
#include <lal/LALStdlib.h>
#include <lal/Units.h>
#include <lal/Random.h>
#include <lal/AVFactories.h>
#include <lal/Date.h>
#include <lal/TimeDelay.h>
#include <lal/DetectorSite.h>
#include <lal/Comm.h>

/* LAL MPI interface */
#include <lal/Comm.h>

/* Associated headers */
#include <ExtTrig.h>
#include <ExtTrigConditionData.h>

#ifdef __cplusplus
extern "C" {
#endif

NRCSID( EXTTRIGCONDITIONDATAC, "$Id: ExtTrigConditionData.c,v 1.4 2002/04/30 17:25:40 rrahkola Exp $");

static void 
GetTimeSeriesIndex( 
		   LALStatus          *status, 
		   UINT4              *index,
		   LIGOTimeGPS        *gpsRef, 
		   const multiDimData *md,
		   const REAL8         delay  );

static void
AddGPSTimes(
	    LALStatus         *status,
	    LIGOTimeGPS       *output,
	    const LIGOTimeGPS *const input1,
	    const LIGOTimeGPS *const input2 );

static void
FindSegments(
	     LALStatus             *status,
	     ETSegList             *output,
	     TwoDetsTimeAndASource *const config,
	     const ETInitParams    *const params );

/************************* wrapperAPI function definitions  ******************/

/*
 * Function LALConditionData
 *
 * Reads the input data provided to the wrapperAPI from LDAS (usually
 * from the data-conditioning API) which is a list of sequences.
 * The code is looking for time-series data with the name "H(L)1(2):LSC-AS_Q".
 *
 */
void
LALConditionData(
    LALStatus      *status,
    LALSearchInput *inout,
    void           *searchParams,
    LALMPIParams   *mpiParams    )
{
  UINT4                 i,j;            /* iterator(s)                    */
  /* variables for communicating between master/slave */
  MPIMessage            mpiMsg;         /* send/rec'v messages w/slave    */
  UINT4                 mpiSize = 0;    /* number of parallel nodes       */
  /* variables for assigning multiDimData */
  INT4                  chan1Index= -1; /* identify chan1 in multiDimData */
  INT4                  chan2Index= -1; /* identify chan2 in multiDimData */
  //INT4                  dq1Index  = -1; /* identify chan1 data quality    */
  //INT4                  dq2Index  = -1; /* identify chan2 data quality    */
  /* variables for filling the Segment list */
  TwoDetsTimeAndASource *tmpConfig =    /* for calculating delays         */
    (TwoDetsTimeAndASource *) LALMalloc( sizeof(TwoDetsTimeAndASource) );
  LALPlaceAndGPS       *detTime1;       /* temp pointer                   */
  LALPlaceAndGPS       *detTime2;       /* temp pointer                   */
  /* variables for filling the DataSegment vector */
  CHARVector           *selVector =NULL;/* fills 'selected' parameters    */
  UINT4                 tsIndexStart, tsIndexEnd; /* for extracted ts     */
  LIGOTimeGPS           gpsDelay;       /* for 2nd timeseries epoch start */
  LIGOTimeGPS           adjGPSStart;    /* for 2nd timeseries epoch start */
  /* pointers for various places in the searchParams structure */
  ETSearchParams       *params    = (ETSearchParams *) searchParams;
  ETInitParams         *initParams= params->initParams;
  ETDataSegment        *dataSeg   = params->data->data;
  ETSegment            *segments  = params->segments->data;
  /* variables for picking random segments and setting random delays */
  static RandomParams  *randGen;                  /* random number        */
#ifdef DEBUG_EXTTRIG
  INT4                  seed      = 12345;        /* 0 = use current time */
#else
  INT4                  seed      = 0;            /* 0 = use current time */
#endif // DEBUG_EXTTRIG
  REAL4                 randNext  = 0.0;/* the actual randomly-generated #*/
  UINT4                 randSeg   = 0;  /* a random index of segList      */
  UINT4                 randLeft  = initParams->numOffSegments;
  /* temporary variables for constructing an optimal filter */
  /* :NOTE: optimal filter construction objects to follow */
  COMPLEX8             *filtData  = NULL;
  COMPLEX8              unity;
  UINT4                 filtLength= initParams->segLength;


  INITSTATUS( status, "LALConditionData", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  /*
   * check that searchParams is valid and cast it to ETSearchParams*
   */
  ASSERT( searchParams, status, CONDITIONDATAH_ENULL, 
	  CONDITIONDATAH_MSGENULL );
  params = (ETSearchParams *) searchParams;

  /*
   * some initial Node-specific code here
   */
  /* Determines whether we're running as master (0) or as a slave (>0) */
  MPI_Comm_rank( *(mpiParams->comm), &(params->rank) );
  /* get the number of nodes */
  MPI_Comm_size( *(mpiParams->comm), &(mpiSize));
  mpiSize -= 1;     /* subtract 1 to get the number of slaves */
#ifdef DEBUG_EXTTRIG
  if( params->searchMaster )
    snprintf( params->id, IDLEN, "master[node %2d]", params->rank );
  else
      snprintf( params->id, IDLEN, "slave[node %2d]", params->rank );
  fprintf(stdout, "exttrig:LALConditionData- %s\n", params->id );
  fprintf( stdout, "%s: params->rank = %d,\tmpiSize = %d\n", 
	   params->id, params->rank, mpiSize                );
  fflush(stdout);
#endif // DEBUG_EXTTRIG
  
  /*
   *  check the input
   */

  /* check that inout data exists */
  ASSERT( inout, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  /* check that the number of multiDimData sequences (time series, frequency
   * series, time-frequency series, database entries, etc.) received from the 
   * datacondAPI is correct.  With two h(t) channels and one/two data quality 
   * channels, this could be three or four.
   * :NOTE: currently only using the two h(t) channels.
   */
  if ( inout->numberSequences != 2 )
    {
      ABORT( status, CONDITIONDATAH_ENSEQ, CONDITIONDATAH_MSGENSEQ );
    }

  /* check that the multidimData sequence data exists */
  if ( !(inout->sequences) )
    {
      ABORT( status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );
    }

  /*
   *
   * allocate temporary memory
   *
   */
  /* optimal filter
   * :NOTE: Current filter is set for unity gain at all frequencies,
   * :NOTE:  non-optimized
   */
  unity.re = 1.0;
  unity.im = 0.0;
  filtData = (COMPLEX8 *) LALMalloc( filtLength * sizeof(COMPLEX8) );
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Optimal filter length = %d\n", params->id,filtLength );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
  for ( i = 0; i < filtLength; i++ )
    {
      filtData[i] = unity;
    }
  /* CHARVector for segment selection
   */
  LALCHARCreateVector( status->statusPtr, &selVector, 
		       params->segments->length );
  CHECKSTATUSPTR( status );
  for (i = 0; i < selVector->length; i++)
    selVector->data[i] = 0; /* make sure everything's cleared out first */
  /* RandomParams RandNext for generating random numbers
   */
  LALCreateRandomParams( status->statusPtr, &randGen, seed );
  CHECKSTATUSPTR( status );

  /* tmpConfig (needs two LALPlaceAndGPS structures)
   */
  detTime1 = tmpConfig->p_det_and_time1 = (LALPlaceAndGPS *) 
    LALMalloc( sizeof(LALPlaceAndGPS) );
  detTime2 = tmpConfig->p_det_and_time2 = (LALPlaceAndGPS *) 
    LALMalloc( sizeof(LALPlaceAndGPS) );
  /* tentatively point to NULL, but we'll change it in FindSegments() */
  tmpConfig->p_source = NULL;

  /* first LALPlaceAndGPS structure
   */
  detTime1->p_detector = (LALDetector *) LALMalloc( sizeof(LALDetector) );
  /* temporarily point to NULL, but we'll change it in FindSegments() */
  detTime1->p_gps = NULL;

  /* second LALPlaceAndGPS structure
   */
  detTime2->p_detector = (LALDetector *) LALMalloc( sizeof(LALDetector) );
  /* permanently point to NULL, since it won't be used while calculating 
     time delay */
  detTime2->p_gps = NULL;


  /*
   *
   * identify the multiDimData sequences by their names (specified in the Tcl
   * script), and check that the data has the correct type, dimensions, and
   * length
   *
   */

  for ( i=0; i < inout->numberSequences; ++i )
    {
      /* check for a name match that also matches the type of data (timeD,
       * freqD, bothD, or databaseD)
       */
      if ( strstr (inout->sequences[i].comment, params->chan1) &&
	   (inout->sequences[i].space == timeD) )
	{
	  /* assign an index number for reference later on */
	  chan1Index = i;
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning chan1 to multiDimData %d: %s\n", 
		   params->id, i, inout->sequences[i].comment );
	  fprintf( stdout, "chan1 multiDimData stats: numberDimensions = %d\n",
		   inout->sequences[i].numberDimensions );
	  fflush (stdout);
#endif // DEBUG_EXTTRIG > 1

	  /* check dimensions (should be 1 for a single REAL4 time series) */
	  if ( inout->sequences[i].numberDimensions != 1 )
	    {
	      ABORT( status, CONDITIONDATAH_EDIMS, CONDITIONDATAH_MSGEDIMS );
	    }

	  /* check type of data (need REAL4) */
	  if ( inout->sequences[i].type != real_4 )
	    {
	      ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
	    }

	  /* :NOTE: Don't actually check length now, but should in the future*/
	  
	}
      else if ( strstr (inout->sequences[i].comment, params->chan2) &&
		(inout->sequences[i].space == timeD) )
	{
	  /* assign an index number for reference later on */
	  chan2Index = i;
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: Assigning chan2 to multiDimData %d: %s\n",
		   params->id, i, inout->sequences[i].comment );
	  fprintf( stdout, "chan2 multiDimData stats: numberDimensions = %d\n",
		   inout->sequences[i].numberDimensions );
	  fflush (stdout);
#endif

	  /* check dimensions */
	  if ( inout->sequences[i].numberDimensions != 1 )
	    {
	      ABORT (status, CONDITIONDATAH_EDIMS, CONDITIONDATAH_MSGEDIMS );
	    }

	  /* check type of data (need REAL4) */
	  if ( inout->sequences[i].type != real_4 )
	    {
	      ABORT( status, CONDITIONDATAH_ETYPE, CONDITIONDATAH_MSGETYPE );
	    }

	  /* :NOTE: Don't actually check length now, but should in the future*/

	}
      /* :TODO: Handle data quality multiDimData for both IFOs */
    }

  /*
   * check that all necessary multiDimData has been assigned
   * :TODO: Need to include data quality channel checks here
   */
  if ( (chan1Index < 0) || (chan2Index < 0) )
    {
      ABORT( status, CONDITIONDATAH_EFIND, CONDITIONDATAH_MSGEFIND );
    }

  /*
   * Fix tmpConfig to use the two IFOs as the LALDetectors
   */
#ifdef DEBUG_EXTTRIG
  /* Use LHO and LLO as det1 and det2, respectively */
  *(detTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
  *(detTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
#else
  /* :NOTE: hard-coding a choice of two detectors is NOT good */
  if( strstr(params->chan1, "H1\\:") || strstr(params->chan1, "H2\\:")   )
    *(detTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
  if( strstr(params->chan2, "H1\\:") || strstr(params->chan2, "H2\\:")   )
    *(detTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLHODIFF];
  if( strstr(params->chan1, "L1\\:") )
    *(detTime1->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
  if( strstr(params->chan2, "L1\\:") )
    *(detTime2->p_detector) = lalCachedDetectors[LALDetectorIndexLLODIFF];
#endif // DEBUG_EXTTRIG
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "Det1: %s,\tDet2: %s\n",
	   detTime1->p_detector->frDetector.name,
	   detTime2->p_detector->frDetector.name );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

  /*
   * Accumulate a list of possible segments
   */
  FindSegments( status->statusPtr, params->segments, tmpConfig, initParams );
  CHECKSTATUSPTR( status );


  /*
   *
   * (Master only): Pick the segments which will be used and pass to slaves
   * (Slaves only): Receive the 'selected' info for the segments
   *
   */
  if( params->searchMaster )
    {
      selVector->data[0] = 1; /* first segment is always selected */
      /* keep picking random numbers until the desired number is obtained,
       * or there are no more segments to pick from
       */
      if (randLeft > params->segments->filled) 
	randLeft = params->segments->filled;
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, "%s: # off-segs:%d,\t# filled:%d,\trandLeft = %d\n",
	       params->id, initParams->numOffSegments,
	       params->segments->filled, randLeft  );
      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
      while (randLeft > 0)
	{
	  LALUniformDeviate( status->statusPtr, &randNext, randGen );
	  CHECKSTATUSPTR( status );
	  randSeg = (INT4)(randNext * (params->segments->length - 1)) + 1;
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, "%s: selecting segment %2d,\trandLeft = %d\n", 
		   params->id, randSeg, randLeft );
	  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
	  if( (randSeg > selVector->length) || selVector->data[randSeg] ) 
	    continue;
	  else
	    {
	      selVector->data[randSeg] = 1;
	      randLeft--;
	    }
	}

      /* now wait for each slave to check in and send them the boolean vector
       */
      while (mpiSize > 0)
	{
	  LALMPIRecvMsg( status->statusPtr, &mpiMsg, *(mpiParams->comm) );
	  CHECKSTATUSPTR( status );
	  if( !mpiMsg.send )
	    {
	      LALMPISendCHARVector( status->statusPtr, selVector, 
				    mpiMsg.source, *(mpiParams->comm) );
	      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
	      fprintf( stdout, 
	               "%s: sent selVector to slave [node %d] \n",
		       params->id, mpiMsg.source );
	      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
	    }
	  else /* shouldn't happen, slaves should be begging for data! */
	    {
	      ABORT( status, CONDITIONDATAH_EMPI, CONDITIONDATAH_MSGEMPI );
	    }
	  mpiSize--;
	}
    }
  else /* must be a slave; just get the boolean vector */
    {
      /* setup the mpiMsg structure 
       */
      mpiMsg.msg    = 0;
      mpiMsg.source = params->rank;
      mpiMsg.send   = 0; /* '0' = will receive data on next MPI exchange(s) */

      /* connect to master and retrieve the vector 
       */
      LALMPISendMsg( status->statusPtr, &mpiMsg, 0, *(mpiParams->comm) );
      CHECKSTATUSPTR( status );
      LALMPIRecvCHARVector( status->statusPtr,selVector,0,*(mpiParams->comm) );
      CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, "%s: rec'vd selVector from master\n",
	       params->id );
      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
    }

  /*
   * fill in the 'selected' parameter for each segment (both master and slave)
   */
  for( i=0; i < selVector->length; i++)/* selVector.size = segments.size */
    segments[i].selected = selVector->data[i];
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, "%s: Selected segments: ", params->id );
  for( i=0; i < selVector->length; i++)
    fprintf( stdout, "%d ", selVector->data[i] );
  fprintf( stdout, "\n" );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1


  /*
   * fill ETDataSegments with h(t) multiDimData and optimal filter data
   * using ETSegList information
   * :NOTE: We're assuming that the data is REAL4. If not, we should
   * :NOTE:  have aborted earlier.
   *
   */
  /* we're going to use two iterators here.  'j' refers to the ETDataSegment,
   * while 'i' refers to the ETSegment we're running over.  We'll pick out
   * time series for any ETSegment[i] which was selected (above), and stick
   * it into ETDataSegment[j], then increment j.
   */
  j = 0;
  params->data->filled = 0;
  params->data->segsToBeAnalyzed = 0;
  for (i = 0; i < params->segments->length; ++i)
    {
      if ( segments[i].selected ) 
	{
	  /*
	   * segmentID
	   */
	  dataSeg[j].segmentID = i;

	  /*
	   * tsChan1
	   */
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexStart), 
			      &(segments[i].startGPS),
			      &(inout->sequences[chan1Index]), 0.0 );
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexEnd), 
			      &(segments[i].endGPS), 
			      &(inout->sequences[chan1Index]), 0.0 );
	  CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, 
     "%s: seg %d; tsChan1 start = %d\ttsChan1 end = %d\t# points = %d,%d\n",
		   params->id, j, tsIndexStart, tsIndexEnd, 
		   (tsIndexEnd - tsIndexStart), initParams->segLength );
	  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

	  memcpy( dataSeg[j].tsChan1->data->data,       /* REAL4 *     */
		  &(inout->sequences[chan1Index].data.real4[tsIndexStart]),
		  dataSeg[j].tsChan1->data->length * sizeof(REAL4) );
	  strncpy( dataSeg[j].tsChan1->name, params->chan1, dbNameLimit );
	  dataSeg[j].tsChan1->epoch = segments[i].startGPS;/*LIGOTimeGPS*/
	  dataSeg[j].tsChan1->deltaT =                  /* REAL8       */
	    inout->sequences[chan1Index].range.dTime.timeStepSize;
	  dataSeg[j].tsChan1->f0 =                      /* REAL8       */
	    inout->sequences[chan1Index].range.dTime.baseFreq;
	  /*
	   * :TODO: Need to read units from the multiDimData
	   */
	  dataSeg[j].tsChan1->sampleUnits = lalADCCountUnit;/* LALUnit */
	  
	  /* 
	   * tsChan2
	   */
	  LALFloatToGPS( status->statusPtr,&(gpsDelay),&(segments[i].delay) );
	  CHECKSTATUSPTR( status );
	  AddGPSTimes( status->statusPtr, &(adjGPSStart),
		       &(segments[i].startGPS), &(gpsDelay) );
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexStart), 
			      &(segments[i].startGPS),
			      &(inout->sequences[chan2Index]),
			      segments[i].delay               );
	  CHECKSTATUSPTR( status );
	  GetTimeSeriesIndex( status->statusPtr, &(tsIndexEnd), 
			      &(segments[i].endGPS), 
			      &(inout->sequences[chan2Index]),
			      segments[i].delay               );
	  CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 1
	  fprintf( stdout, 
      "%s: seg %d; tsChan2 start = %d\ttsChan2 end = %d\t# points = %d,%d\n",
		   params->id, j, tsIndexStart, tsIndexEnd,
		   (tsIndexEnd - tsIndexStart), initParams->segLength );
	  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

	  memcpy( dataSeg[j].tsChan2->data->data,        /* REAL4 *     */
		  &(inout->sequences[chan2Index].data.real4[tsIndexStart]),
		  dataSeg[j].tsChan2->data->length * sizeof(REAL4) );
	  strncpy( dataSeg[j].tsChan2->name, params->chan2, dbNameLimit );
	  dataSeg[j].tsChan2->epoch = adjGPSStart;       /* LIGOTimeGPS */
	  dataSeg[j].tsChan2->deltaT =                   /* REAL8       */
	    inout->sequences[chan2Index].range.dTime.timeStepSize;
	  dataSeg[j].tsChan2->f0 =                       /* REAL8       */
	    inout->sequences[chan2Index].range.dTime.baseFreq;
	  /*
	   * :TODO: Need to read units from the multiDimData
	   */
	  dataSeg[j].tsChan2->sampleUnits = lalADCCountUnit; /* LALUnit */

	  /*
	   * :TODO: Need to copy over data quality information
	   */
	  
	  /*
	   * filter
	   */
	  memcpy( dataSeg[j].filter->data->data, filtData, 
		  filtLength * sizeof(COMPLEX8)       );
	  strncpy( dataSeg[j].filter->name, "Null-effect filter", 
		   LALNameLength * sizeof(CHAR) );
	  dataSeg[j].filter->f0     = 0;
	  dataSeg[j].filter->deltaF = 
	                           (.5 * initParams->segLength) / GW_SAMPLE ;
	  dataSeg[j].filter->sampleUnits= lalHertzUnit;

	  /* increment j,data->filled, and data->segsToBeAnalyzed
	   *when successful 
	   */
	  j++;
	  params->data->filled++;
	  params->data->segsToBeAnalyzed++;
	} /* if ( segments[i].selected )...              */
    } /* for ( i = 0; i < params->segments->length; i++ )... */
      
  /* free temporary allocated memory */
  LALFree( detTime2->p_detector );
  LALFree( detTime1->p_detector );
  LALFree( detTime2 );
  LALFree( detTime1 );
  LALDestroyRandomParams( status->statusPtr, &randGen );
  CHECKSTATUSPTR( status );
  LALCHARDestroyVector( status->statusPtr, &selVector );
  CHECKSTATUSPTR( status );
  LALFree( filtData );
  filtData = NULL;
  LALFree( tmpConfig );
  tmpConfig = NULL;

  /* exit normally */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void 
GetTimeSeriesIndex( 
		   LALStatus          *status, 
		   UINT4              *index,
		   LIGOTimeGPS        *gpsRef, 
		   const multiDimData *md,
		   const REAL8         delay  )
{
  REAL8         tsStart;
  REAL8         gpsTime;
  LIGOTimeGPS   multiDimEpoch;

  INITSTATUS( status, "GetTimeSeriesIndex", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  multiDimEpoch.gpsSeconds     = md->range.dTime.startSec;
  multiDimEpoch.gpsNanoSeconds = md->range.dTime.startNan;
  
  LALGPStoFloat( status->statusPtr, &(gpsTime), gpsRef );
  CHECKSTATUSPTR( status );
  gpsTime += delay;

  LALGPStoFloat( status->statusPtr, &(tsStart), &(multiDimEpoch) );
  CHECKSTATUSPTR( status );
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "GetTimeSeriesIndex: gpsTime = %20.9f\n", gpsTime );
  fprintf( stdout, "                    tsStart = %20.9f\n", tsStart );
  fprintf( stdout, "               timeStepSize = %20.9f\n", 
	   md->range.dTime.timeStepSize );
  fprintf( stdout, "(gpsTime - tsStart) / timeStepSize = %d\n",
	   (UINT4) ((gpsTime - tsStart) / md->range.dTime.timeStepSize) );
  fflush( stdout );
#endif  // DEBUG_EXTTRIG > 2

  if ( tsStart > gpsTime )
    {
      ABORT( status, CONDITIONDATAH_ETIME, CONDITIONDATAH_MSGETIME );
    }

  *index = (UINT4) ( (gpsTime - tsStart) / md->range.dTime.timeStepSize );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void
AddGPSTimes(
	    LALStatus         *status,
	    LIGOTimeGPS       *output,
	    const LIGOTimeGPS *const input1,
	    const LIGOTimeGPS *const input2 ) {
  INT4 tmpGPSSec  = input1->gpsSeconds + input2->gpsSeconds;
  INT4 tmpGPSNSec = input1->gpsNanoSeconds + input2->gpsNanoSeconds;
  INT4 tmpQuot = 0; /* "quotient" for carry-over from nanoseconds to seconds */

  INITSTATUS( status, "AddGPSTimes", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  if( input1->gpsSeconds < 0 ) tmpGPSNSec -= (2 * input1->gpsNanoSeconds);
  if( input2->gpsSeconds < 0 ) tmpGPSNSec -= (2 * input2->gpsNanoSeconds);
  if( abs(tmpGPSNSec) > 1000000000 ) tmpQuot = tmpGPSNSec / 1000000000 ;
  else if( tmpGPSNSec < 0 && tmpGPSSec > 0 ) tmpQuot = -1;
  else if( tmpGPSNSec > 0 && tmpGPSSec < 0 ) tmpQuot = 1;

  output->gpsSeconds     = tmpGPSSec + tmpQuot;
  output->gpsNanoSeconds = abs(tmpGPSNSec - (tmpQuot * 1000000000));
  
#if DEBUG_EXTTRIG > 2
  fprintf( stdout, "AddGPSTimes: %9d.%9d + %9d.%9d = %9d.%9d\n", 
	   input1->gpsSeconds, input1->gpsNanoSeconds,
	   input2->gpsSeconds, input2->gpsNanoSeconds,
	   output->gpsSeconds, output->gpsNanoSeconds );
  fflush( stdout );
#endif  // DEBUG_EXTTRIG > 2

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

static void
FindSegments(
	     LALStatus             *status,
	     ETSegList             *output,
	     TwoDetsTimeAndASource *const config,
	     const ETInitParams    *const params ) {
  /* variables for filling the Segments list */
  UINT4                 i;              /* iterator(s)                    */
  LIGOTimeGPS           tmpGPS;         /* create a GPS time for a delay  */
  SkyPosition           tmpSource;      /* for calculating delays         */
  ETSegment            *segments = output->data;
  LALPlaceAndGPS       *detTime1 = config->p_det_and_time1;/*temp pointer */
  //LALPlaceAndGPS       *detTime2 = config->p_det_and_time2;/*temp pointer */

  INITSTATUS( status, "FindOffSegments", EXTTRIGCONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  /*
   *
   * Accumulate a list of possible segments ('on'-source == index 0)
   *
   */

  /* initial ('on'-source) segment
   */
  /* set the start GPS time of the segment */
  tmpGPS.gpsSeconds     = -(params->onTimeShift);
  tmpGPS.gpsNanoSeconds = 0;
  AddGPSTimes( status->statusPtr, &(segments[0].startGPS),
		      &(params->grbStart), &tmpGPS        );
  CHECKSTATUSPTR( status );
  /* set the end GPS time of the segment */
  tmpGPS.gpsSeconds    += params->segLength / GW_SAMPLE;
  AddGPSTimes( status->statusPtr, &(segments[0].endGPS),
	       &(params->grbStart), &tmpGPS             );
  CHECKSTATUSPTR( status );
  /* set the delay */
  config->p_source = &(params->sourceDir);
  detTime1->p_gps = &(segments[0].startGPS);
  config->p_det_and_time2->p_gps = &(segments[0].startGPS);
  LALTimeDelay( status->statusPtr, &(segments[0].delay), config );
  CHECKSTATUSPTR( status );
  /* 'on'-source segments are always selected! */
  segments[0].selected = 1;
#if DEBUG_EXTTRIG > 1
  fprintf( stdout, 
	   "seg%2d-- startGPS: %d.%d;\tendGPS: %d.%d;\tdelay: %6.5f\n", 0,
	   segments[0].startGPS.gpsSeconds, 
	   segments[0].startGPS.gpsNanoSeconds,
	   segments[0].endGPS.gpsSeconds,
	   segments[0].endGPS.gpsNanoSeconds,
	   segments[0].delay                 );
  fflush( stdout );
#endif // DEBUG_EXTTRIG > 1
  output->filled = 1; /* keep track of how many segments are filled */
		
  /* rest of the ('off'-source) segments
   */
  tmpGPS.gpsSeconds     = (params->offTimeShiftLow); /* initialize tmpTime */
  for( i = 1; i < output->length; i++ )
    {
      /* set the start GPS time of the segment */
      AddGPSTimes( status->statusPtr, &(segments[i].startGPS),
		   &(params->grbStart), &tmpGPS               );
      CHECKSTATUSPTR( status );
      /* set the end GPS time of the segment */
      tmpGPS.gpsSeconds += params->segLength / GW_SAMPLE;
      AddGPSTimes( status->statusPtr, &(segments[i].endGPS),
		   &(params->grbStart), &tmpGPS             );
      CHECKSTATUSPTR( status );
      /* set the delay */
      /* :TODO: include these next lines, but change to random delay
      config.pSource = &(params->sourceDir);
      detTime1->pGPS = &(segments[i].startGPS);
      LALTimeDelay( status->statusPtr, segments[i].delay, config );
      CHECKSTATUSPTR( status );
      */
      segments[i].delay    = 0.0; /* temporarily set to 0 */
      segments[i].selected = 0;
#if DEBUG_EXTTRIG > 1
      fprintf( stdout, 
	       "seg%2d-- startGPS: %d.%d;\tendGPS: %d.%d;\tdelay: %6.5f\n", i,
	       segments[i].startGPS.gpsSeconds, 
	       segments[i].startGPS.gpsNanoSeconds,
	       segments[i].endGPS.gpsSeconds,
	       segments[i].endGPS.gpsNanoSeconds,
	       segments[i].delay                    );
      fflush( stdout );
#endif // DEBUG_EXTTRIG > 1

      /* increment tmpTime by 1 second to put some space between segments */
      tmpGPS.gpsSeconds++;
      /* increment the number of filled segments */
      output->filled++;
    }		

  /* exit normally */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}

#ifdef __cplusplus
}
#endif
